@@ -5,10 +5,12 @@ from collections import MutableMapping
 from typing import get_type_hints, Union
 from dotenv import load_dotenv
 
-load_dotenv()
 
-_DATABASE_URI = 'postgresql://datapusher_jobs:thepassword@localhost/datapusher_jobs'
-_WRITE_ENGINE_URL = 'postgresql://datapusher:thepassword@localhost/datastore_default'
+env_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), '.env')
+load_dotenv(env_file)
+
+_DATABASE_URI = 'postgresql://datapusher_jobs:YOURPASSWORD@localhost/datapusher_jobs'
+_WRITE_ENGINE_URL = 'postgresql://datapusher:YOURPASSWORD@localhost/datastore_default'
 _TYPES = 'String', 'Float', 'Integer', 'DateTime', 'Date', 'NULL'
 _TYPE_MAPPING = {
     'String': 'text', 'Integer': 'numeric', 
@@ -41,13 +43,14 @@ class DataPusherPlusConfig(MutableMapping):
     SSL_VERIFY: bool = False
     TYPES: tuple = _TYPES
     TYPE_MAPPING: dict = _TYPE_MAPPING
-    LOG_FILE: str = '/tmp/ckan_service.log'
+    LOG_FILE: str = '/etc/ckan/datapusher-plus/ckan_service.log'
     STDERR: bool = True
     QSV_BIN: str = '/usr/local/bin/qsvdp'
     PREVIEW_ROWS: int = 1000
     QSV_DEDUP: bool = True
     DEFAULT_EXCEL_SHEET: int = 0
     AUTO_ALIAS: bool = True
+    AUTO_ALIAS_UNIQUE: bool = False
     WRITE_ENGINE_URL: str = _WRITE_ENGINE_URL
     DOWNLOAD_PROXY: str = ''
 
@@ -108,7 +111,6 @@ class DataPusherPlusConfig(MutableMapping):
 # Expose config object for app to import
 config = DataPusherPlusConfig(os.environ)
 
-
 # Expose these two variables so ckanserviceprovider can use it
 SQLALCHEMY_DATABASE_URI = config.get('SQLALCHEMY_DATABASE_URI')
 WRITE_ENGINE_URL = config.get('WRITE_ENGINE_URL')
