@@ -46,16 +46,6 @@ POSTGRES_INT_MIN = -2147483648
 POSTGRES_BIGINT_MAX = 9223372036854775807
 POSTGRES_BIGINT_MIN = -9223372036854775808
 
-# if a field has any of these anywhere in their name, date inferencing
-# is turned on when scanning for data types, which is a relatively expensive op
-# if DATELIKE_FIELDNAMES is empty, date inferencing will always be on
-_DATELIKE_FIELDNAMES = ['date', 'time', 'open', 'close', 'due']
-
-DATELIKE_FIELDNAMES = web.app.config.get(
-    'DATELIKE_FIELDNAMES', _DATELIKE_FIELDNAMES)
-
-DATELIKE_FIELDNAMES = [field.lower() for field in DATELIKE_FIELDNAMES]
-
 DATASTORE_URLS = {
     'datastore_delete': '{ckan_url}/api/action/datastore_delete',
     'resource_update': '{ckan_url}/api/action/resource_update'
@@ -411,7 +401,7 @@ def push_to_datastore(task_id, input, dry_run=False):
     tmp.seek(0)
 
     if (resource.get('hash') == file_hash and not data.get('ignore_hash')):
-        logger.info("Upload skipped as the file hash hasn't changed: {hash}.".format(
+        logger.warning("Upload skipped as the file hash hasn't changed: {hash}.".format(
             hash=file_hash))
         return
 
@@ -430,8 +420,12 @@ def push_to_datastore(task_id, input, dry_run=False):
             qsv_input_csv.close()
         if 'qsv_dedup_csv' in globals():
             qsv_dedup_csv.close()
+        if 'qsv_headers' in globals():
+            qsv_headers.close()
         if 'qsv_safenames_csv' in globals():
             qsv_safenames_csv.close()
+        if 'qsv_applydp_csv' in globals():
+            qsv_applydp_csv.close()
 
     '''
     Start Analysis using qsv instead of messytables, as 1) its type inferences are bullet-proof
@@ -443,6 +437,8 @@ def push_to_datastore(task_id, input, dry_run=False):
         DataSize(length), fetch_elapsed))
     analysis_start = time.perf_counter()
 
+    qsv_bin = config.get('QSV_BIN')
+
     # check content type or file extension if its a spreadsheet
     spreadsheet_extensions = ['XLS', 'XLSX', 'ODS', 'XLSM', 'XLSB']
     format = resource.get('format').upper()
@@ -464,35 +460,57 @@ def push_to_datastore(task_id, input, dry_run=False):
         default_excel_sheet = config.get('DEFAULT_EXCEL_SHEET')
         try:
             qsv_excel = subprocess.run(
-                [config.get('QSV_BIN'), 'excel', qsv_spreadsheet.name, '--sheet', str(default_excel_sheet),
+                [qsv_bin, 'excel', qsv_spreadsheet.name, '--sheet', str(default_excel_sheet),
                  '--trim', '--output', qsv_excel_csv.name],
                 check=True, capture_output=True, text=True)
         except subprocess.CalledProcessError as e:
             cleanup_tempfiles()
-            raise util.JobError(
-                'Cannot export spreadsheet to CSV: {}'.format(e)
-            )
+            logger.error('Upload aborted. Cannot export spreadsheet to CSV: {}'.format(e))
+            return
         qsv_spreadsheet.close()
         excel_export_msg = qsv_excel.stderr
         logger.info("{}...".format(excel_export_msg))
         tmp = qsv_excel_csv
     else:
         '''
-        its a CSV/TSV, not a spreadsheet. Check if its RFC4180 valid & transcode to UTF-8
-        at the same time. If its a TSV/TAB file, convert to CSV as well for standard handling
-        downstream note that we only change the workfile, the resource file itself is unchanged
+        its a CSV/TSV/TAB file, not a spreadsheet. Do two-stage validation: 
+        
+        1) Normalize & transcode to UTF-8 using `qsv input`. We need to normalize as it could be
+        a CSV/TSV/TAB dialect with differing delimiters, quoting, etc. 
+        Using qsv input's --output option also auto-transcodes to UTF-8.
+        
+        2) Run an RFC4180 check with `qsv validate` against the normalized, UTF-8 encoded CSV file.
+        If it passes validation, we can now handle it with confidence downstream as a "normal" CSV.
+        
+        Note that we only change the workfile, the resource file itself is unchanged.
         '''
+        
+        # normalize to CSV, and transcode to UTF-8 if required
         qsv_input_csv = tempfile.NamedTemporaryFile(suffix='.csv')
-        logger.info('Validating/Transcoding {}...'.format(format))
+        logger.info('Normalizing/Transcoding {}...'.format(format))
         try:
             qsv_input = subprocess.run(
-                [config.get('QSV_BIN'), 'input', tmp.name, '--output', qsv_input_csv.name], check=True)
+                [qsv_bin, 'input', tmp.name, '--output', qsv_input_csv.name], check=True)
         except subprocess.CalledProcessError as e:
+            # return as we can't push an invalid CSV file
             cleanup_tempfiles()
-            raise util.JobError(
-                'Invalid CSV file: {}'.format(e)
-            )
+            logger.error("Job aborted as the file cannot be normalized/transcoded: {}.".format(e))
+            return
         tmp = qsv_input_csv
+        logger.info('Normalized & transcoded...')
+        
+        # validation phase
+        logger.info('Validating {}...'.format(format))
+        try:
+            qsv_validate = subprocess.run(
+                [qsv_bin, 'validate', tmp.name], check=True, capture_output=True, text=True)
+        except subprocess.CalledProcessError as e:
+            # return as we can't push an invalid CSV file
+            cleanup_tempfiles()
+            validate_error_msg = qsv_validate.stderr
+            logger.error("Invalid file! Job aborted: {}.".format(validate_error_msg))
+            return
+        logger.info('Valid file...')
 
     # do we need to dedup?
     # note that deduping also ends up creating a sorted CSV
@@ -501,7 +519,8 @@ def push_to_datastore(task_id, input, dry_run=False):
         logger.info('Checking for duplicate rows...')
         try:
             qsv_dedup = subprocess.run(
-                [config.get('QSV_BIN'), 'dedup', tmp.name, '--output', qsv_dedup_csv.name], capture_output=True, text=True)
+                [qsv_bin, 'dedup', tmp.name, '--output', qsv_dedup_csv.name], 
+                capture_output=True, text=True)
         except subprocess.CalledProcessError as e:
             cleanup_tempfiles()
             raise util.JobError(
@@ -514,13 +533,30 @@ def push_to_datastore(task_id, input, dry_run=False):
                 '{:,} duplicates found and removed...'.format(dupe_count))
         else:
             logger.info('No duplicates found...')
+            
+    # get existing header names, so we can use them for data dictionary labels
+    # should we need to change the column name to make it "db-safe"
+    try:
+        qsv_headers = subprocess.run(
+            [qsv_bin, 'headers', '--just-names', tmp.name], 
+            capture_output=True, check=True, text=True)
+    except subprocess.CalledProcessError as e:
+        cleanup_tempfiles()
+        raise util.JobError(
+            'Cannot scan CSV headers: {}'.format(e)
+        )
+    original_headers = str(qsv_headers.stdout).strip()
+    original_header_dict = {idx: ele for idx, ele in 
+                            enumerate(original_headers.splitlines())}
 
-    # ensure our column/header names identifiers are valid postgres identifiers
+    # now, ensure our column/header names identifiers are "safe names"
+    # i.e. valid postgres identifiers
     qsv_safenames_csv = tempfile.NamedTemporaryFile(suffix='.csv')
     logger.info('Checking for safe column names...')
     try:
         qsv_safenames = subprocess.run(
-            [config.get('QSV_BIN'), 'safenames', tmp.name, '--mode', 'verify', '--output', qsv_safenames_csv.name], capture_output=True, text=True)
+            [qsv_bin, 'safenames', tmp.name, '--mode', 'verify', '--output', 
+             qsv_safenames_csv.name], capture_output=True, text=True)
     except subprocess.CalledProcessError as e:
         cleanup_tempfiles()
         raise util.JobError(
@@ -530,76 +566,58 @@ def push_to_datastore(task_id, input, dry_run=False):
     if unsafeheader_count > 0:
         logger.info('Sanitizing header names...')
         qsv_safenames = subprocess.run(
-            [config.get('QSV_BIN'), 'safenames', tmp.name, '--mode', 'conditional', '--output', qsv_safenames_csv.name], capture_output=True, text=True)
+            [qsv_bin, 'safenames', tmp.name, '--mode', 'conditional', 
+             '--output', qsv_safenames_csv.name], 
+            capture_output=True, text=True)
+        tmp = qsv_safenames_csv
     else:
         logger.info('No unsafe header names found...')
-    tmp = qsv_safenames_csv
 
-    # at this stage, we have a "clean" CSV ready for pushing into the datastore
+    # at this stage, we have a "clean" CSV ready for type inferencing
 
     # first, index csv for speed - count, stats and slice
     # are all accelerated/multithreaded when an index is present
     try:
         subprocess.run(
-            [config.get('QSV_BIN'), 'index', tmp.name], check=True)
+            [qsv_bin, 'index', tmp.name], check=True)
     except subprocess.CalledProcessError as e:
         cleanup_tempfiles
         raise util.JobError(
             'Cannot index CSV: {}'.format(e)
         )
 
-    # get record count
+    # get record count, this is instantaneous with an index
     try:
         qsv_count = subprocess.run(
-            [config.get('QSV_BIN'), 'count', tmp.name], capture_output=True, check=True, text=True)
+            [qsv_bin, 'count', tmp.name], capture_output=True, check=True, text=True)
     except subprocess.CalledProcessError as e:
         cleanup_tempfiles()
         raise util.JobError(
             'Cannot count records in CSV: {}'.format(e)
         )
     record_count = int(str(qsv_count.stdout).strip())
+    
+    # its empty, nothing to do
+    if record_count == 0:
+        cleanup_tempfiles()
+        logger.warning('Upload skipped as there are zero records.')
+        return
+    
+    # log how many records we detected
     unique_qualifier = ''
     if config.get('QSV_DEDUP'):
         unique_qualifier = 'unique'
     logger.info('{:,} {} records detected...'.format(
         record_count, unique_qualifier))
 
-    '''
-    if DATELIKE_FIELDNAMES is not empty, scan CSV headers for date-like field,
-    otherwise, always --infer-dates when scanning for types
-    we do this since date type inference is a very expensive op
-    as qsv needs to check if a string is a date using 15 date formats and permutations thereof
-    exponentially slowing down qsv, even if its using SIMD-accelerated, multithreaded regex parsing
-    '''
-    inferdates_flag = True
-    if DATELIKE_FIELDNAMES:
-        try:
-            qsv_headers = subprocess.run(
-                [config.get('QSV_BIN'), 'headers', '--just-names', tmp.name], capture_output=True, check=True, text=True)
-        except subprocess.CalledProcessError as e:
-            cleanup_tempfiles()
-            raise util.JobError(
-                'Cannot scan CSV headers: {}'.format(e)
-            )
-        header_fields = str(qsv_headers.stdout).strip().lower()
-        if not any(datelike_fieldname in header_fields for datelike_fieldname in DATELIKE_FIELDNAMES):
-            inferdates_flag = False
-
     # run qsv stats to get data types and descriptive statistics
     headers = []
     types = []
     headers_min = []
     headers_max = []
     qsv_stats_csv = tempfile.NamedTemporaryFile(suffix='.csv')
-    qsv_stats_cmd = [config.get('QSV_BIN'), 'stats', tmp.name,
-                     '--output', qsv_stats_csv.name]
-    if inferdates_flag:
-        qsv_stats_cmd.append('--infer-dates')
-        # only show the log message when smart date inferencing is on
-        # when smart date inferencing is off, we always infer dates
-        if DATELIKE_FIELDNAMES:
-            logger.info(
-                'Date-like fields detected. Date inferencing enabled...')
+    qsv_stats_cmd = [qsv_bin, 'stats', tmp.name, '--infer-dates', '--dates-whitelist', 
+                     'all', '--output', qsv_stats_csv.name]
     try:
         qsv_stats = subprocess.run(qsv_stats_cmd, check=True)
     except subprocess.CalledProcessError as e:
@@ -631,11 +649,7 @@ def push_to_datastore(task_id, input, dry_run=False):
         }.get(existing_info.get(h, {}).get('type_override'), t)
             for t, h in zip(types, headers)]
 
-    '''
-    Delete existing datastore resource before proceeding. Otherwise
-    'datastore_create' will append to the existing datastore. And if
-    the fields have significantly changed, it may also fail.
-    '''
+    # Delete existing datastore resource before proceeding.
     if existing:
         logger.info('Deleting "{res_id}" from datastore.'.format(
             res_id=resource_id))
@@ -647,6 +661,11 @@ def push_to_datastore(task_id, input, dry_run=False):
                           for field in zip(headers, types)]
 
     # 2nd pass header_dicts, checking for smartint types
+    # and set labels to original column names in case we made the names "db-safe"
+    # as the labels are used by DataTables_view to label columns
+    # we also take note of datetime/timestamp fields, so we can normalize them
+    # to RFC3339 format
+    datetimecols_list = []
     headers_dicts = []
     for idx, header in enumerate(temp_headers_dicts):
         if header['type'] == 'smartint':
@@ -658,7 +677,10 @@ def push_to_datastore(task_id, input, dry_run=False):
                 header_type = 'numeric'
         else:
             header_type = header['type']
-        headers_dicts.append(dict(id=header['id'], type=header_type))
+        if header_type == 'timestamp':
+            datetimecols_list.append(header['id'])
+        info_dict = dict(label=original_header_dict[idx])
+        headers_dicts.append(dict(id=header['id'], type=header_type, info=info_dict))
 
     # Maintain data dictionaries from matching column names
     if existing_info:
@@ -681,7 +703,8 @@ def push_to_datastore(task_id, input, dry_run=False):
         qsv_slice_csv = tempfile.NamedTemporaryFile(suffix='.csv')
         try:
             qsv_slice = subprocess.run(
-                [config.get('QSV_BIN'), 'slice', '--len', str(preview_rows), tmp.name, '--output', qsv_slice_csv.name], check=True)
+                [qsv_bin, 'slice', '--len', str(preview_rows), tmp.name, 
+                 '--output', qsv_slice_csv.name], check=True)
         except subprocess.CalledProcessError as e:
             cleanup_tempfiles()
             raise util.JobError(
@@ -689,11 +712,30 @@ def push_to_datastore(task_id, input, dry_run=False):
             )
         rows_to_copy = preview_rows
         tmp = qsv_slice_csv
+        
+    # if there are any datetime fields, normalize them to RFC3339 format
+    # so we can readily insert them as timestamps into postgresql with COPY
+    if datetimecols_list:
+        qsv_applydp_csv = tempfile.NamedTemporaryFile(suffix='.csv')
+        datecols = ','.join(datetimecols_list)
+        logger.info('Formatting dates \"{}\" to ISO 8601/RFC 3339 format...'.format(datecols))
+        try:
+            qsv_applydp = subprocess.run(
+                [qsv_bin, 'applydp', 'datefmt', datecols, tmp.name, '--output', 
+                qsv_applydp_csv.name], check=True)
+        except subprocess.CalledProcessError as e:
+            cleanup_tempfiles()
+            raise util.JobError(
+                'Applydp error: {}'.format(e)
+            )
+        tmp = qsv_applydp_csv
 
     analysis_elapsed = time.perf_counter() - analysis_start
     logger.info(
         'Analyzed and prepped in {:,.2f} seconds.'.format(analysis_elapsed))
 
+    # at this stage, the resource is ready for COPYing to the Datastore
+
     if dry_run:
         return headers_dicts
 
@@ -762,6 +804,9 @@ def push_to_datastore(task_id, input, dry_run=False):
         resource['preview_rows'] = copied_count
     update_resource(resource, api_key, ckan_url)
 
+    # aliases are human-readable, and make it easier to use than resource id hash
+    # in using the Datastore API and in SQL queries
+    alias_unique_flag = config.get('AUTO_ALIAS_UNIQUE')
     if config.get('AUTO_ALIAS'):
         # get package info, so we can construct the alias
         package = get_package(resource['package_id'], ckan_url, api_key)
@@ -773,24 +818,27 @@ def push_to_datastore(task_id, input, dry_run=False):
         if owner_org:
             owner_org_name = owner_org.get('name')
         if resource_name and package_name and owner_org_name:
+            # we limit it to 59, so we still have space for sequence suffix
+            # postgres max identifier length is 63
             alias = f"{resource_name}-{package_name}-{owner_org_name}"[:59]
-            # check if the alias already exist, if it does
+            # if AUTO_ALIAS_UNIQUE is true, check if the alias already exist, if it does
             # add a sequence suffix so the new alias can be created
-            cur.execute('SELECT COUNT(*) FROM _table_metadata where name like \'{}%\';'.format(
-                alias))
-            alias_count = cur.fetchone()[0]
-            if alias_count:
-                alias_sequence = alias_count + 1
-                while True:
-                    # we do this, so we're certain the new alias does not exist
-                    # just in case they deleted an older alias with a lower sequence #
-                    alias = f'{alias}-{alias_sequence:03}'
-                    cur.execute('SELECT COUNT(*) FROM _table_metadata where name like \'{}%\';'.format(
-                        alias))
-                    alias_exists = cur.fetchone()[0]
-                    if not alias_exists:
-                        break
-                    alias_sequence += 1
+            if alias_unique_flag:
+                cur.execute('SELECT COUNT(*) FROM _table_metadata where name like \'{}%\';'.format(
+                    alias))
+                alias_count = cur.fetchone()[0]
+                if alias_count:
+                    alias_sequence = alias_count + 1
+                    while True:
+                        # we do this, so we're certain the new alias does not exist
+                        # just in case they deleted an older alias with a lower sequence #
+                        alias = f'{alias}-{alias_sequence:03}'
+                        cur.execute('SELECT COUNT(*) FROM _table_metadata where name like \'{}%\';'.format(
+                            alias))
+                        alias_exists = cur.fetchone()[0]
+                        if not alias_exists:
+                            break
+                        alias_sequence += 1
         else:
             logger.info(
                 'Cannot create alias: {}-{}-{}'.format(resource_name, package_name, owner_org))
@@ -800,6 +848,11 @@ def push_to_datastore(task_id, input, dry_run=False):
     # tell CKAN to calculate_record_count and set alias if set
     send_resource_to_datastore(resource, headers_dicts, api_key, ckan_url,
                                records=None, aliases=alias, calculate_record_count=True)
+    
+    # TODO: Create indices automatically based on statistics
+    # e.g. columns with Cardinality = rowcount has all unique values and a unique index can be created
+    # and if CREATE_INDEX setting is set
+    
     if alias:
         logger.info('Created alias: {}'.format(alias))
 
